@@ -1,4 +1,4 @@
-distance <- function(x, method="euclidean", sprange=NULL, spweight=NULL)
+distance <- function(x, method="euclidean", sprange=NULL, spweight=NULL, icov)
 {
 # calculates similarity and dissimilarity coefficients
 # as described in Legendre and Legendre 1998
@@ -39,22 +39,27 @@ distance <- function(x, method="euclidean", sprange=NULL, spweight=NULL)
 # 9: Modified Gower base 10 (Anderson et al 2006)
 # 10: Modified Gower base 2 (Anderson et al 2006)
 
+### Mahalanobis distance requires covariance matrix of full dataset,
+# making it impossible to calculate in chunks for a very large dataset.
+# Now can specify covariance matrix separately, to enable this. 
+# argument icov. If missing, will be calculated from data as in previous version.
+
 pairedsum <- function(x)
 {
-	### paired sums
+    ### paired sums
    ### returns an N by N by P matrix containing each
    ### combination of 
     N <- nrow(x)
-	 P <- ncol(x)
-	 A <- numeric(N * N * P)
+     P <- ncol(x)
+     A <- numeric(N * N * P)
     A <- .C("psum",
-		as.double(as.vector(t(x))),
-		as.integer(N),
-		as.integer(P),
-		A = as.double(A),
-		PACKAGE = "ecodist")$A
-	 
-	 A <- array(A, dim=c(N, N, P))
+        as.double(as.vector(t(x))),
+        as.integer(N),
+        as.integer(P),
+        A = as.double(A),
+        PACKAGE = "ecodist")$A
+     
+     A <- array(A, dim=c(N, N, P))
     A
 }
 
@@ -62,16 +67,16 @@ paireddiff <- function(x)
 {
 ### paired differences
     N <- nrow(x)
-	 P <- ncol(x)
-	 A <- numeric(N * N * P)
+     P <- ncol(x)
+     A <- numeric(N * N * P)
     A <- .C("pdiff",
-		as.double(as.vector(t(x))),
-		as.integer(N),
-		as.integer(P),
-		A = as.double(A),
-		PACKAGE = "ecodist")$A
-	 
-	 A <- array(A, dim=c(N, N, P))
+        as.double(as.vector(t(x))),
+        as.integer(N),
+        as.integer(P),
+        A = as.double(A),
+        PACKAGE = "ecodist")$A
+     
+     A <- array(A, dim=c(N, N, P))
     A
 }
 
@@ -79,16 +84,16 @@ jointpresence <- function(x)
 {
 ### joint count of presences
     N <- nrow(x)
-	 P <- ncol(x)
-	 A <- numeric(N * N * P)
+     P <- ncol(x)
+     A <- numeric(N * N * P)
     A <- .C("jpres",
-		as.double(as.vector(t(x))),
-		as.integer(N),
-		as.integer(P),
-		A = as.double(A),
-		PACKAGE = "ecodist")$A
-	 
-	 A <- array(A, dim=c(N, N, P))
+        as.double(as.vector(t(x))),
+        as.integer(N),
+        as.integer(P),
+        A = as.double(A),
+        PACKAGE = "ecodist")$A
+     
+     A <- array(A, dim=c(N, N, P))
     A
 }
 
@@ -96,16 +101,16 @@ jointabsence <- function(x)
 {
 ### joint count of absences
     N <- nrow(x)
-	 P <- ncol(x)
-	 A <- numeric(N * N * P)
+     P <- ncol(x)
+     A <- numeric(N * N * P)
     A <- .C("jabs",
-		as.double(as.vector(t(x))),
-		as.integer(N),
-		as.integer(P),
-		A = as.double(A),
-		PACKAGE = "ecodist")$A
-	 
-	 A <- array(A, dim=c(N, N, P))
+        as.double(as.vector(t(x))),
+        as.integer(N),
+        as.integer(P),
+        A = as.double(A),
+        PACKAGE = "ecodist")$A
+     
+     A <- array(A, dim=c(N, N, P))
     A
 }
 
@@ -113,16 +118,16 @@ firstonly <- function(x)
 {
 ### present only in first sample
     N <- nrow(x)
-	 P <- ncol(x)
-	 A <- numeric(N * N * P)
+     P <- ncol(x)
+     A <- numeric(N * N * P)
     A <- .C("jfirst",
-		as.double(as.vector(t(x))),
-		as.integer(N),
-		as.integer(P),
-		A = as.double(A),
-		PACKAGE = "ecodist")$A
-	 
-	 A <- array(A, dim=c(N, N, P))
+        as.double(as.vector(t(x))),
+        as.integer(N),
+        as.integer(P),
+        A = as.double(A),
+        PACKAGE = "ecodist")$A
+     
+     A <- array(A, dim=c(N, N, P))
     A
 }
 
@@ -130,16 +135,16 @@ secondonly <- function(x)
 {
 ### present only in second sample
     N <- nrow(x)
-	 P <- ncol(x)
-	 A <- numeric(N * N * P)
+     P <- ncol(x)
+     A <- numeric(N * N * P)
     A <- .C("jsec",
-		as.double(as.vector(t(x))),
-		as.integer(N),
-		as.integer(P),
-		A = as.double(A),
-		PACKAGE = "ecodist")$A
-	 
-	 A <- array(A, dim=c(N, N, P))
+        as.double(as.vector(t(x))),
+        as.integer(N),
+        as.integer(P),
+        A = as.double(A),
+        PACKAGE = "ecodist")$A
+     
+     A <- array(A, dim=c(N, N, P))
     A
 }
 
@@ -154,7 +159,7 @@ x <- as.matrix(x)
     if (method == -1) 
         stop("ambiguous distance method")
     N <- nrow(x)
-	 P <- ncol(x)
+     P <- ncol(x)
 
 
 if(method == 1) 
@@ -162,7 +167,7 @@ if(method == 1)
 # Euclidean distance
     A <- paireddiff(x)
     if(is.null(spweight)) {
-	    D <- sqrt(apply(A, 1:2, function(x)sum(x * x)))
+        D <- sqrt(apply(A, 1:2, function(x)sum(x * x)))
     }
     else if(spweight[1] == "absence") {
         W <- ifelse(jointabsence(x)==1, 0, 1)
@@ -181,9 +186,9 @@ if(method == 2)
 {
 # Bray-Curtis distance
      A <- paireddiff(x)
-	 A <- apply(A, 1:2, function(x)sum(abs(x)))
-	 B <- pairedsum(x)
-	 B <- apply(B, 1:2, sum)
+     A <- apply(A, 1:2, function(x)sum(abs(x)))
+     B <- pairedsum(x)
+     B <- apply(B, 1:2, sum)
      D <- A / B
 }
 
@@ -193,7 +198,7 @@ if(method == 3)
 
    A <- paireddiff(x)
     if(is.null(spweight)) {
-	    D <- apply(A, 1:2, function(x)sum(abs(x)))
+        D <- apply(A, 1:2, function(x)sum(abs(x)))
     }
     else if(spweight[1] == "absence") {
         W <- ifelse(jointabsence(x)==1, 0, 1)
@@ -212,22 +217,25 @@ if(method == 4)
 {
 # pairwise Mahalanobis distance
 # same as mahal()
-	icov <- solve(cov(x))
-	A <- paireddiff(x)
-	A1 <- apply(A, 1, function(z)(z %*% icov %*% t(z)))
-	D <- A1[seq(1, N*N, by=(N+1)), ]
+    if(missing(icov)) {
+        icov <- cov(x)
+    }
+    icov <- solve(icov)
+    A <- paireddiff(x)
+    A1 <- apply(A, 1, function(z)(z %*% icov %*% t(z)))
+    D <- A1[seq(1, N*N, by=(N+1)), ]
 }
 
 if(method == 5)
 {
 # Jaccard distance
     A <- jointpresence(x)
-	 A <- apply(A, 1:2, sum)
-	 B <- firstonly(x)
-	 B <- apply(B, 1:2, sum)
-	 C <- secondonly(x)
-	 C <- apply(C, 1:2, sum)
-	 D <- 1 - A / (A + B + C)
+     A <- apply(A, 1:2, sum)
+     B <- firstonly(x)
+     B <- apply(B, 1:2, sum)
+     C <- secondonly(x)
+     C <- apply(C, 1:2, sum)
+     D <- 1 - A / (A + B + C)
 }
 
 if(method == 6)
@@ -240,27 +248,29 @@ if(method == 7)
 {
 # Sorensen distance
     A <- jointpresence(x)
-	 A <- apply(A, 1:2, sum)
-	 B <- firstonly(x)
-	 B <- apply(B, 1:2, sum)
-	 C <- secondonly(x)
-	 C <- apply(C, 1:2, sum)
-	 D <- 1 - (2*A) / (2*A + B + C)
+     A <- apply(A, 1:2, sum)
+     B <- firstonly(x)
+     B <- apply(B, 1:2, sum)
+     C <- secondonly(x)
+     C <- apply(C, 1:2, sum)
+     D <- 1 - (2*A) / (2*A + B + C)
 }
 if(method == 8) 
 {
     # Gower distance
     # weighting
     A <- paireddiff(x)
-    if(!is.null(sprange) & length(sprange) == ncol(x)) {
-        sprange <- array(rep(sprange, each=nrow(x)^2), dim=c(nrow(x), nrow(x), ncol(x)))
-        A <- A / sprange
-    }
-    else {
-        stop("sprange not recognized.\n")
+    if(!is.null(sprange)) {
+        if(length(sprange) == ncol(x)) {
+            sprange <- array(rep(sprange, each=nrow(x)^2), dim=c(nrow(x), nrow(x), ncol(x)))
+            A <- A / sprange
+        }
+        else {
+            stop("sprange not recognized.\n")
+        }
     }
     if(is.null(spweight)) {
-	    D <- apply(A, 1:2, function(x)sum(abs(x)))
+        D <- apply(A, 1:2, function(x)sum(abs(x)))
     }
     else if(spweight[1] == "absence") {
         W <- ifelse(jointabsence(x)==1, 0, 1)
@@ -284,7 +294,7 @@ if(method == 9)
 
     A <- paireddiff(x)
     if(is.null(spweight)) {
-	    D <- apply(A, 1:2, function(x)sum(abs(x)))
+        D <- apply(A, 1:2, function(x)sum(abs(x)))
     }
     else if(spweight[1] == "absence") {
         W <- ifelse(jointabsence(x)==1, 0, 1)
@@ -307,7 +317,7 @@ if(method == 10)
 
     A <- paireddiff(x)
     if(is.null(spweight)) {
-	    D <- apply(A, 1:2, function(x)sum(abs(x)))
+        D <- apply(A, 1:2, function(x)sum(abs(x)))
     }
     else if(spweight[1] == "absence") {
         W <- ifelse(jointabsence(x)==1, 0, 1)
@@ -324,14 +334,14 @@ if(method == 10)
 
 
 
-## Make the results lower triangular	 
+## Make the results lower triangular     
     D <- D[col(D) < row(D)]
 
 ## give the results attributes similar to dist()
     attr(D, "Size") <- N
-	 attr(D, "Labels") <- rownames(x)
-	 attr(D, "Diag") <- FALSE
-	 attr(D, "Upper") <- FALSE
+     attr(D, "Labels") <- rownames(x)
+     attr(D, "Diag") <- FALSE
+     attr(D, "Upper") <- FALSE
     attr(D, "method") <- METHODS[method]
     class(D) <- "dist"
     D
